/* eslint-env jest */
const ModECS = require('../index')

describe('ModECS', () => {

    let engine

    beforeEach(() => {
        engine = ModECS()
        engine.registerComponent('POSITION', {x:0,y:0})
    })

    it('should create a component', () => {
        expect(() => engine.createComponent('POSITION')).not.toThrow()
    })

    it('should create a component with values', () => {
        const position = engine.createComponent('POSITION', {x:2,y:2})

        expect(position.x).toEqual(2)
        expect(position.y).toEqual(2)
    })

    it('should create an entity', () => {
        const id = engine.createEntity()
        expect(id).toBe(0)
    })

    it('should add registered component to an entity', () => {
        const id = engine.createEntity()
        engine.addEntity(id)
        expect(() => engine.addComponent(id, 'POSITION')).not.toThrow()
    })
})

describe('Scenario 1', () => {
    const engine = ModECS()

    engine.registerComponent('POSITION', {x:0,y:0})
    engine.registerComponent('TARGET', {x:0,y:0})


    engine.registerSystem(
        'Move',
        ['POSITION','TARGET'],
        () => (position, target, id) => {
            position.x += target.x
            position.y += target.y
        }
    )

    // engine.compile()

    const e1 = engine.createEntity()
    engine.addEntity(e1)
    engine.addComponent(e1, 'POSITION')
    engine.addComponent(e1, 'TARGET')
    
    const e2 = engine.createEntity()
    engine.addEntity(e2)
    engine.addComponent(e2, 'POSITION')
    engine.addComponent(e2, 'TARGET')


    it('should', () => {
        const moveView = engine.createView('POSITION')

        expect(moveView.POSITION).toHaveLength(2)


        engine.removeComponent(e1, 'POSITION')

        expect(moveView.POSITION).toHaveLength(2)

        engine.updateComponent(e2, 'POSITION', {x:1,y:1})
        engine.updateComponent(e2, 'TARGET', {x:11,y:11})


        const moveView2 = engine.createView('POSITION', 'TARGET')

        engine.process()

        expect(moveView2.POSITION).toHaveLength(1)

        

    })

})

describe('Scenario 2', () => {
    const engine = ModECS()

    engine.registerComponent('POSITION', {x:0,y:0})
    engine.registerComponent('TARGET', {x:0,y:0})
    engine.registerComponent('ROTATION', {angle:0})
    engine.registerComponent('OTHER', {x:0})


    engine.registerSystem(
        'Move',
        ['POSITION','TARGET'],
        () => (position, target, id) => {
            position.x += target.x
            position.y += target.y
        }
    )

    const e1 = engine.createEntity()
    engine.addEntity(e1)

    engine.addComponent(e1, 'POSITION')
    engine.addComponent(e1, 'TARGET')
    engine.addComponent(e1, 'ROTATION')

    const e2 = engine.createEntity()
    engine.addEntity(e2)

    engine.addComponent(e2, 'POSITION')
    engine.addComponent(e2, 'TARGET')
    engine.addComponent(e2, 'OTHER')


    it('should', () => {
        const moveView = engine.createView('POSITION')
        const targetView = engine.createView('TARGET')

        const customView = engine.createView('POSITION', 'TARGET', 'OTHER')
        const targetRotationView = engine.createView('TARGET','ROTATION')


        const e3 = engine.createEntity()
        engine.addEntity(e3)
        engine.addComponent(e3, 'OTHER')
        engine.addComponent(e3, 'TARGET')
        engine.addComponent(e3, 'OTHER', {x:4})

        
        expect(moveView.POSITION).toHaveLength(2)
        expect(targetView.TARGET).toHaveLength(3)

        expect(customView.POSITION).toHaveLength(1)
        expect(customView.TARGET).toHaveLength(1)
        expect(customView.OTHER).toHaveLength(1)
        
        const e4 = engine.createEntity()
        engine.addEntity(e4)
        engine.addComponent(e4, 'POSITION')
        engine.addComponent(e4, 'TARGET')
        engine.addComponent(e4, 'OTHER', {x:4})

        
        expect(customView.POSITION).toHaveLength(2)
        expect(customView.TARGET).toHaveLength(2)
        expect(customView.OTHER).toHaveLength(2)

        

    })

})

describe('Scenario 3', () => { 
    const engine = ModECS()

    engine.registerComponent('POSITION', {x:0,y:0})
    engine.registerComponent('TARGET', {x:1,y:1})

    engine.registerSystem(
        'Move',
        ['POSITION','TARGET'],
        () => (position, target, id) => {
            position.x += target.x
            position.y += target.y
        }
    )

    // engine.compile()

    const e1 = engine.createEntity()
    engine.addEntity(e1)
    engine.addComponent(e1, 'POSITION')
    engine.addComponent(e1, 'TARGET')

    const e2 = engine.createEntity()
    engine.addEntity(e2)
    engine.addComponent(e2, 'POSITION')
    engine.addComponent(e2, 'TARGET')


    it('should', () => {
        const moveView = engine.createView('POSITION')

        expect(moveView.POSITION).toHaveLength(2)

        engine.removeComponent(e1, 'POSITION')

        engine.process()

        expect(moveView.POSITION).toHaveLength(1)

        engine.updateComponent(e2, 'POSITION', {x:1,y:1})
        engine.updateComponent(e2, 'TARGET', {x:11,y:11})

        const moveView2 = engine.createView('POSITION', 'TARGET')

        expect(moveView2.POSITION).toHaveLength(1)

        const snapshot = engine.snapshot()

        const engine2 = ModECS({snapshot})

        // console.log(engine2.snapshot())

    })

})


describe('Scenario 4', () => {

    const engine = ModECS()

    engine.registerComponent('POSITION', {x:0,y:0})
    engine.registerComponent('TARGET', {x:1,y:1})
    engine.registerComponent('TIME', {tick:0})

    engine.registerSystem(
        'Move',
        ['POSITION','TARGET'],
        () => (position, target, id) => {
            position.x += target.x
            position.y += target.y

            engine.removeComponent(id, 'TARGET')
            engine.addComponent(id, 'TIME', {tick: engine.time.tick})
        }
    )

    const e1 = engine.createEntity()
    engine.addEntity(e1)
    engine.addComponent(e1, 'POSITION')

    const e2 = engine.createEntity()
    engine.addEntity(e2)
    engine.addComponent(e2, 'POSITION')

    it('should', () => {

        engine.addComponent(e1, 'TARGET')

        engine.process()
        engine.time.tick++
        engine.process()
        engine.time.tick++
        engine.process()
        engine.time.tick++

        const pos = engine.getComponent(e1, 'POSITION')
        const time = engine.getComponent(e1, 'TIME')

        expect(pos.x).toBe(1)
        expect(time.tick).toBe(0)

    })

})

describe('Scenario 5', () => {
    const engine = ModECS()
  
    engine.registerComponent('POSITION', {x: 0, y: 0})
    engine.registerComponent('TARGET', {x: 1, y: 1})
  
    const spy = jest.fn(tick => tick)
  
    engine.registerSystem(
          'Move',
          ['POSITION', 'TARGET'],
          () => (position, target, id) => {
            position.x += target.x
            position.y += target.y
  
            if (position.x === 3) {
              engine.removeComponent(id, 'TARGET')
              return
            }
            spy(engine.time.tick)
          }
      )
  
    const e1 = engine.createEntity()
    engine.addEntity(e1)
    engine.addComponent(e1, 'POSITION')
  
    const e2 = engine.createEntity()
    engine.addEntity(e2)
    engine.addComponent(e2, 'POSITION')
  
    it('should', () => {
        engine.addComponent(e1, 'TARGET')

        engine.process()
        engine.time.tick++
        engine.process()
        engine.time.tick++
        engine.process()
        engine.time.tick++
        engine.process()
        engine.time.tick++
        engine.process()
        engine.time.tick++

        const pos = engine.getComponent(e1, 'POSITION')

        expect(pos.x).toBe(3)
        expect(spy.mock.calls.length).toBe(2)
        expect(spy.mock.calls[0][0]).toBe(0)
        expect(spy.mock.calls[1][0]).toBe(1)
        expect(spy.mock.results[1].value).toBe(1)
    })
  })